subroutine df_spec_wv ( npts, dt, xi, xj, tbp, kspec, nf,             &
                           freq, df_spec, df_cohe )

!
!  Modified
!	German Prieto
!	March 2005
!
!  Construct the dual-frequency spectrum, as in 
!  	Mellors, R. Vernon, F. and Thomson, D. J. (1998) 
!  	Detection of dispersive signals using multitaper
!	dual frequency coherence
!	Geophys. J. Int. Vol 135 pp. 146-154
!
!	Note that this subroutine increases the number of
!  	samples of the time series to get an unaliased WV
!  	spectrum if required. Based on df_spectrum.f90 
!

!********************************************************************

   use gutil

   implicit none

!  input

   integer, intent(in) :: npts, nf, kspec
  
   real(8), intent(in) :: dt, tbp

   real(8), intent(in), dimension(npts) :: xi, xj

!  Output

   real(8), intent(out), dimension(nf) :: freq

   real(8), intent(out), dimension(nf,nf) :: df_cohe

   complex(8), intent(out), dimension(nf,nf) :: df_spec


!  spectra and frequency

   real(8), dimension(nf) :: spec_i, spec_j 

   real(8), dimension(nf,kspec) :: wt_i, wt_j

   complex(8), dimension(2*npts,kspec) :: yk_i, yk_j


!  Working matrices

   real(8), dimension(nf,nf)      :: Sii, Sjj
   
   complex(8), dimension(nf)     :: dyk_i, dyk_j, dk_i, dk_j
   complex(8), dimension(nf,nf)   :: dk_ii, dk_jj
   
   complex(8), dimension(nf,nf,kspec) :: dyk_ii, dyk_jj

!  Others

   integer :: i 

   real(8), parameter :: pi = 3.14159265358979d0 

!********************************************************************

!  if xi is xj, then just do one estimate of the spectrum

   if (all(xi==xj)) then    
      call mtspec_wv ( npts,dt,xi,tbp,kspec,nf,freq,          &
                       spec_i,yk_i,wt_i)
      spec_j = spec_i
      yk_j = yk_i
      wt_j = wt_i
   else
      call mtspec_wv ( npts,dt,xi,tbp,kspec,nf,freq,          &
                       spec_i,yk_i,wt_i)

      call mtspec_wv ( npts,dt,xj,tbp,kspec,nf,freq,          &
                       spec_j,yk_j,wt_j)
   endif

!
!  Create the spectra (cannot use spec output, normalized different)
!

   spec_i = sum( wt_i**2 * (abs(yk_i(1:nf,:)))**2 , dim=2)/    &
                     sum((wt_i**2),dim=2)
   spec_j = sum( wt_j**2 * (abs(yk_j(1:nf,:)))**2 , dim=2)/    &
                     sum((wt_j**2),dim=2)

   
!
!  Meshgrid weights and dk*yk
!

   dk_i = sqrt( sum (wt_i(:,:)**2, dim=2) )
   dk_j = sqrt( sum (wt_j(:,:)**2, dim=2) )

   call meshgrid(dk_i, dk_j, dk_ii, dk_jj)

   call meshgrid(spec_i,spec_j,Sii,Sjj)


   do i = 1,kspec
      dyk_i = wt_i(:,i) * yk_i(1:nf,i)
      dyk_j = wt_j(:,i) * yk_j(1:nf,i)
      
      call meshgrid(dyk_i,dyk_j,dyk_ii(:,:,i),dyk_jj(:,:,i))

   enddo

   df_spec  = sum( dyk_ii * conjg(dyk_jj), dim=3 ) / (dk_ii*dk_jj)

   df_cohe  = (abs(df_spec))**2 / (Sii * Sjj)

end subroutine df_spec_wv





