subroutine mtspec_wv (npts,dt,x,tbp,kspec,freq,spec,        	&
		yk, wt )

!
!  Modified
!	German Prieto
!	March 2005
!
!  The subroutine is in charge of estimating the adaptive weigthed 
!  multitaper spectrum, as in Thomson 1982, following the idea
!  in mtspec.f90.
!  Note that this code is specifically designed for use on the 
!  Unaliased Wigner-Ville Spectrum, where the dual-frequency spectrum
!  is Fourier transformed after a 45 degrees rotation. To avoid 
!  aliasing the time series is padded with zeros, after the Prolate
!  tapers are applied, thus doubling frequency resolution, which
!  translates in an unaliased WV spectrum.
!
!  As a by product of the spectrum (spec), intermediate steps 
!  are estimated. By products include the phase information
!  in yk, and the weigths wt(nf,kspec) used.
!
!  Variable names used in the subroutine
!
!   	xmean	mean of the time series (data is demeaned)
!	xvar	variance of time series
!   	df	frequency bin separation 
!	fnyq	Nyquist frequency
!	sk	eigenspectra (yk**2)
!   	yk 	eigencoefficients
!	sbar	mean spectrum
!	se 	number of degrees of freedom for each frequency bin
!	err	jackknife 95% confidence interval
!	wt	array containing the ne weights for kspec eigenspectra 
!		normalized so that the sum of squares over
!           	the kspec eigenspectra is one
!	lambda  eigenvalues of the dpss
!	theta	1 - eigenvalues
!	vn	dpss sequences
!	xi 	Variance efficiency Eq. 7.2(c)
!	seavg	Stability of estimate (should be close to unity) Eq. 5.5
!
!
!  Input
!
!	npts	-	integer number of data points in x
!	dt	- 	real, sampling rate of the time series x
!	x(npts) -	real, data vector
!	tbp	-	real, the time-bandwidth product (NW)
!	kspec	- 	integer, number of tapers to use in the estimation
!
!  Output
!
!	freq(npts)	real vector with frequency bins
!	spec(npts)	real vector with the adaptive estimated spectrum
!   	yk 		eigencoefficients
!	wt(npts)	array containing the ne weights for kspec 
!			eigenspectra normalized so that the sum of squares 
!			over the kspec eigenspectra is one
!
!  calls
!	dpss, eigenft_wv, adaptspec
!

!**********************************************************************

   implicit none

!  Input

   integer, intent(in) :: npts, kspec

   real(4), intent(in) :: dt, tbp

   real(4), intent(in), dimension(npts) :: x

!  Output

   real(4), intent(out), dimension(npts) :: freq

   real(4), intent(out), dimension(npts) :: spec

   real(4), intent(out), dimension(npts,kspec) :: wt

   complex(4), intent(out), dimension(2*(npts+mod(npts,2))-1,kspec) :: yk 


!  Eigenspectra

   real(8), dimension(npts) :: spec8

   real(8), dimension(npts,kspec) :: sk
   
   complex(8), dimension(2*(npts+mod(npts,2))-1,kspec) :: yk8 

!  Simple mean spectra

   real(8), dimension(npts) :: sbar

!  Adaptive spectra

   real(8), dimension(npts) :: se

   real(8), dimension(npts,kspec) :: wt8

   real(8), dimension(npts,2) :: err
   
   
!  Dpss 

   real(8), dimension(kspec) :: lambda, theta
   
   real(8), dimension(npts,kspec) :: vn

!  Efficiency and Stability of estimate

   real(8) :: xi, seavg

!
!  Extras
!

!  Time series

   real(8) :: xmean, xvar 

!  Frequency variables

   real(4) :: df, fnyq

!  Other

   integer :: i, j, nf

   real(8) :: xsum, sscal

   real(8), dimension(npts) :: x2

!********************************************************************

!
! Frequency variables
!

   nf = 2*(npts+mod(npts,2))-1

   fnyq	= 0.5/dt
   df 	= fnyq/real(npts - 1)

!
!  Define frequency bins
!

   do i = 1,npts
      freq(i) = real(i-1)*df
   enddo
  
!
!  Get time series stats (mean and variance)
!  Always demean the time series
!

   xmean = sum(dble(x))/dble(npts)
   xvar = (sum((x - xmean)**2))/dble(npts)

   x2 = dble(x) - xmean

!
!  Get the dpss
!

   call dpss(npts,dble(tbp),kspec,vn,lambda,theta)

   write(6,'(3x,a/(4f18.14))')'Prolate spheroidal eigenvalues:',  &
     			       (lambda(j),j=1,kspec)

!
! Get the eigenspectra (yk's, sk's and smean)
!

   call eigenft_wv(npts, x2, kspec, vn, lambda, nf, yk8, sk, sbar)

!
!  Calcuate adaptive weigthed spectrum
!

   call adaptspec(nf,npts,kspec,theta,yk8, &
                spec8,se,wt8,err)

!
!  double power in positive frequencies
!

   yk = cmplx(yk8)
   wt = real(wt8)

   do i=2,npts
   
      spec8(i) = 2.d0 * spec8(i)
   
   enddo

!
!  scale spectrum to meet parseval's theorem
!
   
   sscal = dble(npts)*xvar/sum(spec8)
      
   spec = real(sscal*spec8)

!
!  Put errors on scale
!

   err(:,1) = spec / err(:,1)
   err(:,2) = spec * err(:,2)
  
!
!  Average stability (average degrees of freedom)
!		(Eq. 5.5 Pg 1065)
!

   seavg=sum(se)/(2.d0*dble(kspec)*dble(npts))
   
   write(6,'(a,g15.7)') ' Average stability of estimate = ', seavg

!
!  compute estimate of variance efficiency 
!		(Eq 7.2(c) Pg 1069)
!
   
   xi=0.
   do i=1,npts
      xsum=0.
      do j=1,kspec
         xsum = xsum+vn(i,j)**2
      enddo
      xsum = xsum/dble(kspec)
      xi = xi+xsum**2
   enddo
      
   xi=1./(dble(npts)*xi)

   write(6,'(a,g15.7)') ' Variance efficiency = ', xi


end subroutine mtspec_wv

