subroutine qiinv(npts,dt,x,tbp,kspec,nf,lambda,vn,yk,wt,  &
                 qispec,err,slope)

!
!  Calculate the Stationary Inverse Theory Spectrum.
!  Basically, compute the spectrum inside the innerband. 
!
!  This approach is very similar to 
!  D.J. Thomson (1990).
!
!  INPUT
!
!	npts		number of points in time series
!	dt		sampling rate
!	x(npts)`	the time series, real
!	tbp		the time-bandwidth product
!	kspec   	number of tapers to use
!	nf		number of freq points (npts/2+1)
!       lambda(kspec)   the eigenvalues of the Slepian sequences
!	vn(npts,kspec)  the slepian sequences
!       yk(npts,kspec)  multitaper eigencoefficients, complex
!	
!  IN/OUT
!
!	wt(nf,kspec)	the weights of the different coefficients. 
!			input is the original multitaper weights, 
!			from the Thomson adaptive weighting. 
!			output are the new QI theory coefficients
!			which minimize the mse using second deriva-
!			tive information. 
!
!  OUTPUT
!
!	qispec(nf)	the QI spectrum
!       err(nf,2)   	real array of jackknife error estimates, 
!			with 5 and 95% confidence intervals of 
!			the spectrum.
!	slope(nf)	the estimate of the first derivative
!
!********************************************************************

   use spectra
   use plot

   implicit none

!  Odd freq points in 2W bandwidth

   integer, parameter                     :: nxi = 79

!  Input

   integer,                           intent(in)   :: kspec, npts,  nf
   real(8),                           intent(in)   :: dt, tbp
   real(8),    dimension(npts),       intent(in)   :: x
   complex(8), dimension(npts,kspec), intent(in)   :: yk 
   real(8),    dimension(kspec),      intent(in)   :: lambda
   real(8),    dimension(npts,kspec), intent(in)   :: vn

!  In/Out & Out

   real(8), dimension(nf,kspec), intent(in out)    :: wt
   real(8), dimension(nf),   intent(out)           :: qispec 
   real(8), dimension(nf,2), intent(out)           :: err
   real(8), dimension(nf),   intent(out)           :: slope

!  Frequency Resampling

   real(8)                                :: bp, df
   real(8)                                :: dxi, ct, st, om
   real(8),    dimension(nxi)             :: rxi
   real(8),    dimension(nxi)             :: xi
   real(8),    dimension(npts)            :: vn_dpss

!  Spectrum parameters

   real(8)                                :: mvar
   real(8),    dimension(kspec)           :: varsk
   real(8),    dimension(nf,kspec)        :: Sk
   complex(8), dimension(nf,kspec)        :: xk

!  Covariance and Projection matrices

   complex(8), dimension(kspec*kspec,nf)  :: C
   complex(8), dimension(nxi,kspec)       :: Vj 
   complex(8), dimension(kspec*kspec,nxi) :: Pk 
   complex(8), dimension(kspec,nxi)       :: Pkk 

!  The Quadratic model

   real(8),    dimension(nf)              :: spec, cte
   real(8),    dimension(nf)              :: quad, freq
   real(8),    dimension(nxi)             :: hcte, hslope, hquad
   complex(8), dimension(3)               :: hmodel
   complex(8), dimension(kspec*kspec,3)   :: hk 

!  The Bias terms (outer and quadratic bias)
 
   real(8),    dimension(kspec)           :: Gk, Bk
   real(8),    dimension(nf,kspec)        :: Bin
   complex(8), dimension(npts,kspec)      :: Vk

!  Spectrum iterations

   real(8)                                :: rerr
   real(8), dimension(nf)                 :: sbar, slast
   real(8), dimension(nf,kspec)           :: Skw 

!  QR Decomposition

   complex(8), dimension(kspec*kspec)     :: btilde
   complex(8), dimension(kspec*kspec,nxi) :: Q
   complex(8), dimension(nxi,nxi)         :: R
   complex(8), dimension(nxi,kspec*kspec) :: Q_T

! Jackknife

   real(4)                                  :: qtdis

   complex(8), dimension(npts,kspec-1)      :: Vk_jack
   real(8),    dimension(kspec-1)           :: lambda_jack
   real(8),    dimension(nf,kspec-1)        :: wt_jack, Sk_jack
   real(8),    dimension(nf,kspec)          :: spec_jack, varjk 

   real(8),    dimension(nf)                :: lspec,lsjk_mean,bjack
   real(8),    dimension(nf)                :: se, var
   real(8),    dimension(nf,kspec)          :: lsjk, wt_dofs

!  Others

   integer                                :: nh, n, ik
   integer                                :: L, i, j, k, m 
   real(8)                                :: wt_sum

   real(8), parameter :: pi=3.141592653589793d0, tpi = 2.d0*pi

!********************************************************************

   if (minval(lambda)<0.9d0) then
      write(6,*) 'Poor leakage properties of eigenvalue ', minval(lambda)
      write(6,*) 'Value of kspec is too large, stopped *****'
      stop
   endif
 
   do k = 1,kspec
      xk(:,k) = wt(:,k)*yk(1:nf,k)
   enddo

!  New inner bandwidth frequency
!  Many of this is double precision.

   bp = tbp/(dble(npts))		! W bandwidth

   dxi = (2.d0*dble(bp))/dble(nxi-1)	! QI freq. sampling

   df = 0.5d0/dble(nf - 1)                ! Assume unit sampling


!   print *, 'bp, dxi, nxi, df ', bp, real(dxi), nxi, df

   xi  = dble( (/(i-(nxi/2)-1, i=1,nxi)/) )*dxi

   do j = 1,kspec
      vn_dpss = vn(:,j)

      do i = 1,nxi

         om = tpi* xi(i) 	! Normalize to unit sampling, for FFT
       
         call sft(vn_dpss,npts,om,ct,st)

         Vj(i,j)  = 1.d0/sqrt(lambda(j)) * cmplx(ct,st,kind=8)

     enddo
   enddo

!   call gplot(xi,real(Vj(:,7)),'hold',color='3')
!   call gplot(xi,real(Vj(:,4)),'hold',color='2')
!   call gplot(xi,real(Vj(:,1)),color='1',output='Vk_real.ps')
!
!   call gplot(xi,dimag(Vj(:,7)),'hold',color='3')
!   call gplot(xi,dimag(Vj(:,4)),'hold',color='2')
!   call gplot(xi,dimag(Vj(:,1)),color='1',output='Vk_imag.ps')

!
!  Create the vectorized Cjk matrix and Pk matrix {Vj Vk*}
!

   L = kspec*kspec

   m = 0 
   do j = 1,kspec
      do k = 1,kspec

         m = m + 1
         C(m,:) = ( conjg(xk(:,j)) * (xk(:,k)) )

         Pk(m,1:nxi) = conjg(Vj(:,j)) * (Vj(:,k))

         if (j==k) then
            Pkk(j,:) = Pk(m,:)
         endif

      enddo
   enddo

   Pk(1:m,1)         = 0.5d0 * Pk(1:m,1)
   Pk(1:m,nxi)       = 0.5d0 * Pk(1:m,nxi)
   Pkk(1:kspec,1)    = 0.5d0 * Pkk(1:kspec,1)
   Pkk(1:kspec,nxi)  = 0.5d0 * Pkk(1:kspec,nxi)

!  Prefered model (constant + slope + quadratic)
!  I use the Chebyshev Polynomial as the expansion basis.

   rxi = dble( (/(i-(nxi/2)-1, i=1,nxi)/) ) / dble(nxi/2) * bp
 
   hcte(1:nxi)  = 1.d0 * bp		
   hk(:,1) = matmul(Pk(1:L,1:nxi),hcte) * dxi


   hslope(1:nxi) = rxi 
   hk(:,2) = matmul(Pk(1:L,1:nxi),hslope) * dxi

   hquad(1:nxi) = (2.d0*((rxi/bp)**2) - 1.d0) * bp 
   hk(:,3) = matmul(Pk(1:L,1:nxi), hquad) * dxi

   open(12,file='hk_files.txt')
   write(12,*) real( hk(:,1))
   write(12,*) aimag(hk(:,1))
   write(12,*) real( hk(:,2))
   write(12,*) aimag(hk(:,2))
   write(12,*) real( hk(:,3))
   write(12,*) aimag(hk(:,3))
   close(12)

   Gk = matmul(Pkk,(hquad+bp)) * (dxi)   ! Maintain units 3 parameters
  
   if (m .ne. L) then 
      write(6,*) 'Error in matrix sizes, stopped '
      stop
   endif 
   n = nxi  
   nh = 3 

!  QR decomposition

   call zqrfac(m,n,hk,Q,R)

   Q_T = transpose(Q)
   Q_T = conjg(Q_T)

   do j = 1,nf

!     The back transformation (for complex variable)

      btilde = matmul(Q_T,C(:,j))

      hmodel = 0.d0

      do i = nh, 1, -1
         do k = i+1, nh
            hmodel(i) = hmodel(i) - R(i,k)*hmodel(k)
         enddo
         hmodel(i) = (hmodel(i) + btilde(i))/R(i,i)
      enddo

      cte(j)   = (hmodel(1))    
      slope(j) = (-hmodel(2))	! To get the correct units (it is switched)
      quad(j)  = (hmodel(3))

   enddo

   freq = dble( (/(i-1, i=1,nf)/) ) * dble(df)
!   call gplot(freq,cte,logxy='linlin',output='s0.ps')
   call gplot(freq(2:nf),slope(2:nf),output='s1.ps',logxy='loglin')
!   call gplot(freq,quad,output='s2.ps')


!  Work with the bias terms

   ! 1. Get the average outer bias terms

   do i = 1,nf
      wt_sum = sum(wt(i,:)**2)/dble(kspec)  ! Each Sk has right units
      wt(i,:) = wt(i,:)/sqrt(wt_sum)        ! sum(Sk)/kspec right units
   enddo

   do k = 1,kspec
!      Sk(:,k) = (abs(yk(1:nf,k)))**2
      Sk(:,k) = (abs(wt(:,k)*yk(1:nf,k)))**2
   enddo

   varsk = (Sk(1,:) + Sk(nf,:) + 2.d0*sum(Sk(2:nf-1,:),1) ) * df
   mvar = sum(varsk)/dble(kspec)


   Bk = (1.d0 - lambda) * mvar 

   ! 2. Get the inner bias, using the second derivative

   do k = 1,kspec
      Bin(:,k) = quad * Gk(k)
   enddo

!  Compute weights and the spectrum

   rerr = 9.5d-7	! The relative improvement on the loops (real(8))

!   sbar = sum(Sk(:,1:2),2)/dble(2) 	! First guess, with 2 first tapers
   sbar = sum(Sk(:,1:kspec),2)/dble(kspec) ! First guess, kspec tapers

   do j = 1,1000

      slast = sbar

      do k = 1,kspec
!         wt(:,k)  = sqrt( lambda(k)*sbar /       &
!                   ( lambda(k)*sbar + abs(Bin(:,k)) + Bk(k)))
!         wt(:,k)  = sqrt( lambda(k)*sbar /       &
!                   ( lambda(k)*sbar + Bk(k)))
         wt(:,k)  = sqrt( sbar /       &
                   ( sbar + abs(Bin(:,k))))
         Skw(:,k) = wt(:,k)**2 * Sk(:,k)
      enddo

      sbar = sum(Skw(:,1:kspec),dim=2) / sum(wt(:,1:kspec)**2, 2)

      if (j==1000) then
         spec = sbar
         write(6,'(a,2E16.7)') 'Did not converge, rerr = ',    &
               maxval(abs((sbar-slast)/(sbar+slast))),rerr
         cycle
      endif
      if (maxval(abs((sbar-slast)/(sbar+slast))) .gt. rerr) then
         cycle
      endif
      
!      write(6,*) maxval(abs((sbar-slast)/(sbar+slast))), ' final', j
      
      qispec = sbar
      exit

   enddo

!  The jackknife estimate of confidence intervals

   do m = 1,kspec
 
      sbar = sum(Sk(:,1:2),2)/dble(2) 	! First guess, with 2 first tapers

      ik = 0
      do i = 1,kspec
         if (i == m) then
            cycle
         endif
         ik = ik + 1
     
         Sk_jack(:,ik)   = Sk(:,i)
         lambda_jack(ik) = lambda(i)
         Vk_jack(:,ik)   = Vk(:,i)
 
      enddo

      do j = 1,1000
 
         slast = sbar

         do k = 1,kspec-1
            wt_jack(:,k)  = sqrt(  lambda_jack(k)*sbar /  &
                            ( lambda_jack(k)*sbar +    &
                              abs(Bin(:,k))+Bk(k)) )

            Skw(:,k) = wt_jack(:,k)**2 * Sk_jack(:,k)

         enddo

         sbar = sum(Skw,dim=2) / sum(wt_jack**2, 2)

         if (j==1000) then
            spec_jack(:,m) = sbar
            write(6,'(a,2E16.7)') 'Did not converge, rerr = ',    &
                  maxval(abs((sbar-slast)/(sbar+slast))),rerr
            cycle
         endif
         if (maxval(abs((sbar-slast)/(sbar+slast))) .gt. rerr) then
            cycle
         endif
      
!         write(6,*) maxval(abs((sbar-slast)/(sbar+slast))), ' jack final', j
      
         spec_jack(:,m) = sbar
         exit

     enddo
   enddo	! end jackknife replicas

!  Jackknife mean (Log S)

   lspec = log(qispec)
   lsjk  = log(spec_jack)

   lsjk_mean = sum(lsjk, dim=2)/dble(kspec)

!  Jackknife Bias estimate (Log S)

   bjack = dble(kspec-1) * (lspec - lsjk_mean)

!  Jackknife Variance Estimate (Log S)

   do i = 1,kspec
      varjk(:,i) = (lsjk(:,i) - lsjk_mean)**2
   enddo
   var = sum(varjk, dim=2) * dble(kspec-1)/dble(kspec)

   do i = 1,nf
      wt_dofs(i,:) = wt(i,:)/sqrt(sum(wt(i,:)**2)/dble(kspec))
   enddo

   wt_dofs = amin1(wt_dofs,1.d0)

   se = 2.d0 * sum(wt_dofs**2, dim=2) 
!   se = 2.d0 * sum(wt, dim=2) 

   do i = 1,nf
      if (se(i)<1.0d0) then
         write(6,'(a,i5,a,f10.5)') 'dof < 1 ', i, 'th frequency',se(i)
	 stop
      endif
      var(i) = exp( dble(qtdis(0.950, real(se(i)) )) *sqrt(var(i)))

   enddo

!  The setting of the confidence interval is done in mtspec.f90.

   err(:,1) = var
   err(:,2) = var

   open(12,file='est_deriv.dat')
   do i = 1,nf
      write(12,*) freq(i), qispec(i), slope(i), quad(i)
   enddo
   close(12)

end subroutine qiinv


