subroutine qiinv(npts,dt,x,tbp,kspec,nf,  &
                 freq,spec,qispec,cte,slope,quad,err)

!
!  Calculate the Stationary Inverse Theory Spectrum.
!  Basically, compute the spectrum inside the innerband. 
!
!  This approach is very similar to 
!  D.J. Thomson (1990).
!
!  INPUT
!
!	npts		number of points in time series
!	dt		sampling rate
!	x(npts)`	the time series, real
!	tbp		the time-bandwidth product
!	kspec   	number of tapers to use
!	nf		number of freq points (npts/2+1)
!
!  OUTPUT
!
!	freq(nf)	frequency vector
!	spec(nf)	the spectrum estimate
!	qispec(nf)	the QI spectrum
!	cte(nf)		the constant estimate
!	slope(nf) 	slope or time derivative
!	quad(nf)	Quadratic term (second derivative)
!
!********************************************************************

   use spectra

   implicit none

!  Odd freq points in 2W bandwidth

   integer, parameter                     :: nxi = 79

!  Input

   integer, intent(in)                    :: kspec, npts,  nf
   real(8), intent(in)                    :: dt, tbp
   real(8), dimension(npts), intent(in)   :: x

!  Out

   real(8), dimension(nf),   intent(out)  :: freq, spec, cte, slope
   real(8), dimension(nf),   intent(out)  :: quad, qispec 
   real(8), dimension(nf,2), intent(out)  :: err

!  Frequency Resampling

   real(8)                                :: bp, df
   real(8)                                :: dxi, ct, st, om
   real(8),    dimension(nxi)             :: rxi
   real(8),    dimension(nxi)             :: xi
   real(8),    dimension(npts)            :: vn_dpss

!  Spectrum parameters

   real(8)                                :: mvar
   real(8),    dimension(kspec)           :: varsk
   real(8),    dimension(nf,kspec)        :: wt, Sk, dk
   complex(8), dimension(nf,kspec)        :: xk
   complex(8), dimension(npts,kspec)      :: yk 

!  Dpss 

   real(8), dimension(kspec)              :: lambda
   real(8), dimension(npts,kspec)         :: vn

!  Covariance and Projection matrices

   complex(8), dimension(kspec*kspec,nf)  :: C
   complex(8), dimension(nxi,kspec)       :: Vj 
   complex(8), dimension(kspec*kspec,nxi) :: Pk 
   complex(8), dimension(kspec,nxi)       :: Pkk 

!  The Quadratic model

   real(8),    dimension(nxi)             :: hcte, hslope, hquad
   complex(8), dimension(3)               :: hmodel
   complex(8), dimension(kspec*kspec,3)   :: hk 

!  The Bias terms (outer and quadratic bias)
 
   real(8),    dimension(kspec)           :: Gk, Bk
   real(8),    dimension(nf,kspec)        :: Bjk, Bin
   real(8),    dimension(npts)            :: freq_bias
   complex(8), dimension(npts)            :: dummy_Vk
   real(8),    dimension(2,npts)          :: dummy_a
   complex(8), dimension(npts,kspec)      :: Vk

!  Spectrum iterations

   real(8)                                :: rerr
   real(8), dimension(nf)                 :: sbar, slast
   real(8), dimension(3*nf-1)             :: Sb
   real(8), dimension(nf,kspec)           :: Skw 

!  QR Decomposition

   complex(8), dimension(kspec*kspec)     :: btilde
   complex(8), dimension(kspec*kspec,nxi) :: Q
   complex(8), dimension(nxi,nxi)         :: R
   complex(8), dimension(nxi,kspec*kspec) :: Q_T

! Jackknife

   real(4)                                  :: qtdis

   real(8),    dimension(nf,kspec-1)        :: Bjk_jack
   complex(8), dimension(npts,kspec-1)      :: Vk_jack
   real(8),    dimension(kspec-1)           :: lambda_jack
   real(8),    dimension(nf,kspec-1)        :: wt_jack, Sk_jack
   real(8),    dimension(nf,kspec)          :: spec_jack, varjk 

   real(8),    dimension(nf)                :: lspec,lsjk_mean,bjack
   real(8),    dimension(nf)                :: se, var
   real(8),    dimension(nf,kspec)          :: lsjk, wt_dofs

!  Others

   integer                                :: nh, n, p, ik
   integer                                :: L, i, j, k, m 
   integer, dimension(npts)               :: indx 
   integer, dimension(3*nf-1)             :: sindx

   real(8), parameter :: pi=3.141592653589793d0, tpi = 2.d0*pi

!********************************************************************

!  Get the spectrum estimate

   call mtspec(npts,dt,x,tbp,kspec,nf,freq,spec,   &
                yk=yk, wt=dk, vn=vn,lambda=lambda  )
 
   if (minval(lambda)<0.9d0) then
      write(6,*) 'Poor leakage properties of eigenvalue ', minval(lambda)
      write(6,*) 'Value of kspec is too large, stopped *****'
      stop
   endif
 
   do k = 1,kspec
!      xk(:,k) = yk(1:nf,k)
      xk(:,k) = dk(:,k)*yk(1:nf,k)
   enddo

!  New inner bandwidth frequency
!  Many of this is double precision.

   bp = tbp/(dble(npts))		! W bandwidth

   dxi = (2.d0*dble(bp))/dble(nxi-1)	! QI freq. sampling

   df = 0.5d0/dble(nf - 1)                ! Assume unit sampling


   print *, 'bp, dxi, nxi, df ', bp, real(dxi), nxi, df

   xi  = dble( (/(i-(nxi/2)-1, i=1,nxi)/) )*dxi

   do j = 1,kspec
      vn_dpss = vn(:,j)

      do i = 1,nxi

         om = tpi* xi(i) 	! Normalize to unit sampling, for FFT
       
         call sft(vn_dpss,npts,om,ct,st)

         Vj(i,j)  = 1.d0/sqrt(lambda(j)) * cmplx(ct,st,kind=8)

     enddo
   enddo

!
!  Create the vectorized Cjk matrix and Pk matrix {Vj Vk*}
!

   L = kspec*kspec

   m = 0 
   do j = 1,kspec
      do k = 1,kspec

         m = m + 1
         C(m,:) = ( conjg(xk(:,j)) * (xk(:,k)) )

         Pk(m,1:nxi) = conjg(Vj(:,j)) * (Vj(:,k))

         if (j==k) then
            Pkk(j,:) = Pk(m,:)
         endif

      enddo
   enddo

   Pk(1:m,1)         = 0.5d0 * Pk(1:m,1)
   Pk(1:m,nxi)       = 0.5d0 * Pk(1:m,nxi)
   Pkk(1:kspec,1)    = 0.5d0 * Pkk(1:kspec,1)
   Pkk(1:kspec,nxi)  = 0.5d0 * Pkk(1:kspec,nxi)

!  Prefered model (constant + slope + quadratic)
!  I use the Chebyshev Polynomial as the expansion basis.

   rxi = dble( (/(i-(nxi/2)-1, i=1,nxi)/) ) / dble(nxi/2) * bp
 
   hcte(1:nxi)  = 1.d0 * bp		
   hk(:,1) = matmul(Pk(1:L,1:nxi),hcte) * dxi


   hslope(1:nxi) = rxi 
   hk(:,2) = matmul(Pk(1:L,1:nxi),hslope) * dxi

   hquad(1:nxi) = (2.d0*((rxi/bp)**2) - 1.d0) * bp 
   hk(:,3) = matmul(Pk(1:L,1:nxi), hquad) * dxi

   Gk = matmul(Pkk,(hquad+bp)) * (dxi)   ! Maintain units 3 parameters
  
   print *, 'nxi = ' , nxi, rxi(1), rxi(nxi), ((nxi/2))

   if (m .ne. L) then 
      write(6,*) 'Error in matrix sizes, stopped '
      stop
   endif 
   n = nxi  
   nh = 3 

!  QR decomposition

   call zqrfac(m,n,hk,Q,R)

   Q_T = transpose(Q)
   Q_T = conjg(Q_T)

   do j = 1,nf

!     The back transformation (for complex variable)

      btilde = matmul(Q_T,C(:,j))

      hmodel = 0.d0

      do i = nh, 1, -1
         do k = i+1, nh
            hmodel(i) = hmodel(i) - R(i,k)*hmodel(k)
         enddo
         hmodel(i) = (hmodel(i) + btilde(i))/R(i,i)
      enddo

      cte(j)   = (hmodel(1))    
      slope(j) = (hmodel(2))
      quad(j)  = (hmodel(3))

   enddo

!  Work with the bias terms

   ! 1. Get the average outer bias terms

   do k = 1,kspec
      Sk(:,k) = (abs(yk(1:nf,k)))**2
   enddo

   varsk = (Sk(1,:) + Sk(nf,:) + 2.d0*sum(Sk(2:nf-1,:),1) ) * df
   mvar = sum(varsk)/dble(kspec)

   print *, 'QI variance ', mvar

   Bk = (1.d0 - lambda) * mvar 

   print *, 'QI Bk ', Bk

   ! 2. Get the inner bias, using the second derivative

   do k = 1,kspec
      Bin(:,k) = quad * Gk(k)
   enddo

   ! 3. Generate the Vk(f) windows

   p = ceiling(real(npts)/2.d0)

   indx = (/(p+i, i=1,npts-p), (i, i=1,p)/)    ! the FFT shift

   do k = 1,kspec
      dummy_a(1,:) = vn(:,k)
      dummy_a(2,:) = 0.d0

      dummy_Vk = cmplx(dummy_a(1,:),dummy_a(2,:),kind=8)
      
      call fft(dummy_Vk,npts)
     
      do i = 1,npts 
         Vk(i,k) = dummy_Vk(indx(i))
      enddo

   enddo

   freq_bias = dble( (/(i-(npts/2)-1, i=1,npts)/) ) / dble(npts/2) * 0.5d0 

   do i = 1,npts
      if (abs(freq_bias(i)) <= bp ) then
         Vk(i,:) = cmplx(0.d0,0.d0,kind=8)
      endif
   enddo

   if (mod(npts,2)==1) then 		! odd number of points
      sindx = (/(nf-(i-1), i=1,nf-1), (i, i=1,nf), (nf-(i-1), i=1,nf)/)
   else					! even number of points
      sindx = (/(nf-(i-1), i=1,nf-1), (i, i=1,nf), (nf-i, i=1,nf)/)
   endif 

!  Compute weights and the spectrum

   rerr = 9.5d-10	! The relative improvement on the loops (real(8))

   sbar = sum(Sk(:,1:2),2)/dble(2) 	! First guess, with 2 first tapers
!   sbar = sum(Sk(:,1:kspec),2)/dble(kspec) ! First guess, kspec tapers

   do j = 1,1000
 
      slast = sbar

      do i = 1,3*nf-1
         Sb(i) = sbar(sindx(i))
      enddo
    
      do i = 1,nf		! Get better estimate of outer bias
         do k = 1,kspec
            Bjk(i,k) = (sum( Sb(i:i-1+npts) *abs(Vk(:,k))**2 )) *df 
            !Bjk(i,k) = max(Bjk(i,k),Bk(k))
         enddo
      enddo

      do k = 1,kspec
!         wt(:,k)  = sqrt( lambda(k)*dble(sbar) /       &
!                   ( lambda(k)*dble(sbar) + abs(Bin(:,k)) + Bjk(:,k))) 
         wt(:,k)  = sqrt( lambda(k)*sbar /       &
                   ( lambda(k)*sbar + abs(Bin(:,k)) + Bk(k)))
         Skw(:,k) = wt(:,k)**2 * Sk(:,k)
      enddo

      sbar = sum(Skw(:,1:kspec),dim=2) / sum(wt(:,1:kspec)**2, 2)

      if (j==1000) then
         spec = sbar
         write(6,'(a,2E16.7)') 'Did not converge, rerr = ',    &
               maxval(abs((sbar-slast)/(sbar+slast))),rerr
         cycle
      endif
      if (maxval(abs((sbar-slast)/(sbar+slast))) .gt. rerr) then
         cycle
      endif
      
      write(6,*) maxval(abs((sbar-slast)/(sbar+slast))), ' final', j
      
      qispec = sbar
      exit

   enddo

   qispec = qispec * dt

!  The jackknife estimate of confidence intervals

   do m = 1,kspec
 
      sbar = sum(Sk(:,1:2),2)/dble(2) 	! First guess, with 2 first tapers

      ik = 0
      do i = 1,kspec
         if (i == m) then
            cycle
         endif
         ik = ik + 1
     
         Sk_jack(:,ik)   = Sk(:,i)
         lambda_jack(ik) = lambda(i)
         Vk_jack(:,ik)   = Vk(:,i)
 
      enddo

      do j = 1,1000
 
         slast = sbar

         do i = 1,3*nf-1
            Sb(i) = sbar(sindx(i))
         enddo
    
         do i = 1,nf		! Get better estimate of outer bias
            do k = 1,kspec-1
               Bjk_jack(i,k) = (sum(Sb(i:i-1+npts)    &
                               *abs(Vk_jack(:,k))**2 )) *df 
            enddo
         enddo

         do k = 1,kspec-1
            wt_jack(:,k)  = sqrt(  lambda_jack(k)*sbar /  &
                            ( lambda_jack(k)*sbar +    &
                              abs(Bin(:,k))+Bjk_jack(:,k)) )

            Skw(:,k) = wt_jack(:,k)**2 * Sk_jack(:,k)

         enddo

         sbar = sum(Skw,dim=2) / sum(wt_jack**2, 2)

         if (j==1000) then
            spec_jack(:,m) = sbar
            write(6,'(a,2E16.7)') 'Did not converge, rerr = ',    &
                  maxval(abs((sbar-slast)/(sbar+slast))),rerr
            cycle
         endif
         if (maxval(abs((sbar-slast)/(sbar+slast))) .gt. rerr) then
            cycle
         endif
      
!         write(6,*) maxval(abs((sbar-slast)/(sbar+slast))), ' jack final', j
      
         spec_jack(:,m) = sbar
         exit

     enddo
   enddo	! end jackknife replicas

!  Jackknife mean (Log S)

   lspec = log(qispec)
   lsjk  = log(spec_jack)

   lsjk_mean = sum(lsjk, dim=2)/dble(kspec)

!  Jackknife Bias estimate (Log S)

   bjack = dble(kspec-1) * (lspec - lsjk_mean)

!  Jackknife Variance Estimate (Log S)

   do i = 1,kspec
      varjk(:,i) = (lsjk(:,i) - lsjk_mean)**2
   enddo
   var = sum(varjk, dim=2) * dble(kspec-1)/dble(kspec)

   do i = 1,nf
      wt_dofs(i,:) = wt(i,:)/sqrt(sum(wt(i,:)**2)/dble(kspec))
   enddo

   wt_dofs = amin1(wt_dofs,1.d0)

!   se = 2.d0 * sum(wt_dofs**2, dim=2) 
   se = 2. * sum(wt**2, dim=2) 

   do i = 1,nf
      if (se(i)<1.0) then
         write(6,'(a,i5,a,f10.5)') 'dof < 1 ', i, 'th frequency',se(i)
	 stop
      endif
      var(i) = exp( dble(qtdis(0.950, real(se(i)) )) *sqrt(var(i)))

   enddo

!  The setting of the confidence interval is done in mtspec.f90.

   err(:,1) = var
   err(:,2) = var

   do i = 14,47
      print *, real(freq(i)), real(dk(i,1)), real(wt(i,1)), & 
                              real(dk(i,2)), real(wt(i,2)), &
                              real(dk(i,5)), real(wt(i,5)), &
                              real(dk(i,6)), real(wt(i,6)), &
                              real(dk(i,7)), real(wt(i,7))
   enddo

   print *, 'The two bias terms'
   do i = 14,47
      print *, real(freq(i)), real(Bjk(i,1)), real(Bin(i,1)), & 
                              real(Bjk(i,2)), real(Bin(i,2)), &
                              real(Bjk(i,5)), real(Bin(i,5)), &
                              real(Bjk(i,6)), real(Bin(i,6)), &
                              real(Bjk(i,7)), real(Bin(i,7)), real(qispec(i)/dt)
   enddo

   print *, 'High frequencies'
   do i = nf-5,nf
      print *, real(freq(i)), real(dk(i,1)), real(wt(i,1)), & 
                              real(dk(i,2)), real(wt(i,2)), &
                              real(dk(i,5)), real(wt(i,5)), &
                              real(dk(i,6)), real(wt(i,6)), &
                              real(dk(i,7)), real(wt(i,7))
   enddo
   print *, 'The two bias terms'
   do i = nf-5,nf
      print *, real(freq(i)), real(Bjk(i,1)), real(Bin(i,1)), & 
                              real(Bjk(i,2)), real(Bin(i,2)), &
                              real(Bjk(i,5)), real(Bin(i,5)), &
                              real(Bjk(i,6)), real(Bin(i,6)), &
                              real(Bjk(i,7)), real(Bin(i,7)), real(qispec(i)/dt)
   enddo


end subroutine qiinv


