subroutine wv_spec ( npts,dt,x,tbp,kspec,                   &
                     filter, fbox, filename )  
           

!
!  Construct the Wigner-Ville spectrum from the dual-frequency 
!  spectrum, by taking the inverse FFT, to the 45 degrees
!  rotated df_spectrum.
!  Both df_spec and the Wigner-Ville spectrum are save in a file.
!  
!  Filter	0	No filter
!		1	Boxcar filter
!		2	Guassian Filter
!
!  fbox		Width of the filter to use (for boxcar and Gaussian)
!

!********************************************************************

   implicit none

!  Inputs

   integer, intent(in) :: npts, kspec, filter

   real(4), intent(in) :: dt, tbp, fbox

   real(4), dimension(npts), intent(in) :: x

   character (len=100), intent(in) :: filename

!  spectra and frequency

   integer :: nf, nf2 

   real(4), dimension(npts) :: spec

   real(4), dimension(npts,kspec) :: wt 

   complex(4), dimension(2*(npts+mod(npts,2))-1,kspec) :: yk

!  spectra and frequency

   real(4), dimension(npts) :: freq

!  Dual freq matrices

   complex(4), dimension(npts,kspec)     :: dyk 
   real(4), dimension(npts)              :: dk 
   
   real(4), dimension(npts)    :: df_cohe
 
   complex(4), dimension(npts) :: df_spec

!  Others

   integer :: i, j, k

   character (len = 100) :: fmt, fmt2, file1, file2

   real(4), parameter :: pi = 3.14159265358979 

!  Matrix rotation

   integer :: indices, ncol, gmax

   complex(4), dimension(npts) :: x2

!  Filtering matrix

   real(4) :: sigma
   real(4), dimension(npts) :: x_filt, i_filt, df_filt


!********************************************************************

   nf = 1 + npts/2
   nf2 = npts
   
   write(6,'(a,2i5)') 'Data points and frequency points', npts, nf2

!
!  Get the spectrum estimate
!

   call mtspec_wv ( npts,dt,x,tbp,kspec,freq,          &
                    spec,yk,wt)
      
!
!  Create the spectra (cannot use spec output, normalized different)
!

   spec = sum( wt**2 * (abs(yk(1:nf2,:)))**2 , dim=2)/    &
                     sum((wt**2),dim=2)

   dk = sqrt( sum (wt(:,:)**2, dim=2) )

   dyk(:,:) = wt(:,:) * yk(1:nf2,:)

!  Filter vector

   i_filt = real( (/((i-1), i=1,nf2)/) )

   sigma  = real(nf2)/fbox

   if (filter==2) then
   
      write(6,'(a)') 'Applying Gaussian filter'
      
   ! Guassian Curve

      x_filt = exp( real(-1./2.) *     &
               (sigma * (i_filt-real(nf2/2))/real(nf2))**2);
      df_filt = exp( real(-1./2.) *    &
                (sigma * (i_filt/2.)/real(nf2))**2);

   elseif (filter==1) then
 
      write(6,'(a)') 'Applying Boxcar filter'

   ! Boxcar filter
  
      do i =1,nf2
      
         if (abs(i_filt(i)-real(nf2/2)) > fbox) then
            x_filt(i) = 0.
         else
            x_filt(i) = 1.
         endif

         if (i_filt(i)/2. > fbox) then
            df_filt(i) = 0.
         else
            df_filt(i) = 1.
         endif

      enddo

   else 
   
   ! Flat

      x_filt  = 1.
      df_filt = 1.

   endif

   file1 = trim(filename) // '_wv.dat'
   file2 = trim(filename) // '_df.dat'

   open(12,file=file1,form='formatted')
 
   open(13,file=file2,form='formatted')

   write(fmt,'(i12)') npts
   fmt = '(' // trim(adjustl(fmt)) // 'E16.7)'

   write(fmt2,'(i12)') nf2
   fmt2 = '(' // trim(adjustl(fmt2)) // 'E16.7)'

   do i=1,nf2

      if (mod(i,1000) == 0) then
         write(6,'(a,i5,a,i5)') 'Loop ',i , ' of ', nf2 
      endif

   ! df_spectrum and coherence
   
      do j = 1,nf2
         
         df_spec(j) = sum ( dyk(i,:) * conjg(dyk(j,:)) )   &
                            / (dk(i)*dk(j))

         df_cohe(j) = (abs(df_spec(j)))**2 / (spec(i)*spec(j))

	 ! Filter

         k = nint( sign(0.5,real(j-i))-sign(0.5,-real(j-i)) ) * (j-i) +1 

         df_cohe(j) = df_cohe(j) * df_filt(k)

      enddo
      
      write(13,fmt2) ( real(df_cohe(:)) )

   ! Wigner-Vlle

      gmax = min(i-1,nf2-i,nint(nf2/2.)-1)

      ncol = i
      
      x2 = ( 0., 0. )
      
      do j = -gmax,gmax,1
         
         indices = mod(npts+j,npts) + 1

         x2(indices) = sum ( dyk(ncol+j,:) * conjg(dyk(ncol-j,:)) )   &
                             / (dk(ncol+j)*dk(ncol-j))

         x2(indices) = x2(indices) * x_filt((nf2/2)+1 + j)

      enddo

      call ifft4(x2,npts)

!      if (maxval(imag(x2)/real(x2)) > 1e-8 ) then
!         write(6,'(a)') 'The WV is not real?'
!         write(6,'(2E16.7)')  maxval(imag(x2))
!      endif

  
      write(12,fmt) ( real(x2(:)) )

 
   enddo


   close(12)
   close(13)

end subroutine wv_spec








